// PicWeb.java
//
// This software is Copyright 1998 by Mark Watson.
// This software may be freely used for any purpose
// and may be modified for any purpose as long as
// you agree to the following:
//
// 1. Any bug fixes or enhancements that you make must be emailed
//    back to Mark Watson (markw@markwatson.com) for possible
//    inclusion in the source code base that is available at
//    the web site www.markwatson.com (credit will be given
//    to contributers).
//
// 2. This copyright notice must be kept with the source code
//    and duplicated if you distribute a modified verison of
//    this software in binary form.
//
// 3. You will not re-distribute the source code. Rather, any
//    improvements to this program will be returned to Mark
//    Watson so that there is always one source base at
//    the web site www.markwatson.com (you may freely
//    distribute compiled derived versions of this program
//    without any restrictions as long as this entire
//    copyright notice is reproduced with the application
//    (e.g., in the 'help' or 'about' dialog box).
//
// 4. This software may not be used for any illegal purpose.
//
// 5. You take all responsibility for the use of this software:
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.

// FreeWare contributers:
//
//   Name                        Email (optional)
//   ----                        ----------------
//
//


import java.applet.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.awt.Image;
import java.awt.Graphics;
import java.awt.MediaTracker;
import java.io.*;
import java.net.*;
import java.util.*;

/** PicWeb class contains both the user interface and top level
    logic for the PicWeb Builder application.
*/
public class PicWeb extends Panel implements Serializable, Runnable {

  private int s_pixels[] = new int[256];
  private float [] reds= new float[256];
  private float [] greens= new float[256];
  private float [] blues= new float[256];
  private boolean [] trans= new boolean[256];

  protected boolean showLogo = true;
  protected boolean fileNamesFlag = false;
  protected boolean skipProcessedDirs = false;
  protected Logo logo;
  protected int imageSizeChoice = 1;

  protected Options options;

  public static void main(String args []) {
    AppFrame f = new AppFrame();
    PicWeb pw = new PicWeb();
    f.add(pw);
    f.setTitle("PicWeb 3.0");
    f.pack();
    f.show();
  }

  /** PicWeb class constructor creates the user interface Panel
      and initializes a work thread
  */
  public PicWeb() { 

    DialogFrame o_frame = new DialogFrame();
    options = new Options(o_frame);

    process=false;
    setFont(new Font("Dialog", Font.PLAIN, 12));
    setLayout(null);

    Label l0 = new Label("Picture Web Page Maker(tm) version 3.0.             www.markwatson.com");
    l0.setFont(new Font("Dialog", Font.BOLD, 14));
    add(l0);
    l0.setBounds(5, 3, 510, 30);   
    Label l1 = new Label("root directory:");
    add(l1);
    l1.setBounds(5, 50, 77, 30);
    //root_dir = new TextField(new String(sb) + "test", 40);
    root_dir = new TextField("." + fs + "test", 40);
    add(root_dir);
    root_dir.setBounds(86, 50, 170, 30);

    Button p_button = new Button("Process");
    p_button.addMouseListener(new MouseProcess());
    add(p_button);
    p_button.setBounds(260, 50, 60, 30);
    Button help_button = new Button("Options screen");
    help_button.addMouseListener(new MouseOptions());
    add(help_button);
    help_button.setBounds(325, 50, 95, 30);
    
    logo = new Logo(this, 50, 50);
    add(logo);
    logo.setBounds(430, 30, 50, 50);
    
    outputText = new TextArea(copyright, 6, 82);
    add(outputText);
    outputText.setBounds(5, 100, 500, 110);
    
    setBounds(20, 40, 512, 230);

    workThread = new Thread(this);
    workThread.start();
  }

  /** The inner class MouseOptions is used to detect a mouse
      release event that occurs when a user clicks the 'Options'
      button. The mouseReleased method makes the 'options' object
      visible
  */
  class MouseOptions extends MouseAdapter implements Serializable {
    public void mouseReleased(MouseEvent mevt) {
      options.setVisible(true);
    }
  }

  /** The 'nl' and 'fs' objects are define for string constants
      instead of using literal strings in order to pass the
      "100% Pure" Java certification by Key Labs
  */
  private String nl = System.getProperty("line.separator");
  private String fs = System.getProperty("file.separator");

  /** The 'run' method is called when the work thread is started.
      The work thread sleeps until there is work to be done.  We
      use a separate work thread so that the user interface does not
      "freeze up" when PicWeb is processing image files and generating
      HTML documents.
  */
  public void run() {
    while (true) {
      try {
        workThread.sleep(500);
        if (process) {
          process=false;
          try {
            outputText.setText("");
            ProcessGifs(root_dir.getText());
            outputText.append(nl + "Done processing." + nl);
            logo.clear();
          } catch (IOException e) {
            outputText.setText("caught an io exception in ProcessGifs" + e.toString());
          }
        }
      } catch (Exception e) { }
    }
  }

  /** The inner MouseProcess class is used to detect mouse released
      events in the 'Process' button. The boolean flag 'process' is
      set to 'true' which causes the 'run' method to process a
      directory tree
  */
  class MouseProcess extends MouseAdapter implements Serializable {
    public void mouseReleased(MouseEvent mevt) {
      process = true;;
    }
  }

  private Image output_image = null;

  /** Method ProcessGifs processes an entire directory tree by
      calling itself recursively
  */
  final private void ProcessGifs(String directory) throws IOException {

    outputText.append("processing:" + directory + nl);
    File dir = new File(directory);
        
    if (!dir.isDirectory()) 
      throw new IllegalArgumentException("no such directory");
    File cwd = dir;
    
    String[] entries = cwd.list(new GraphicsFileFilter());
    
    int num = entries.length;
    
    if (skipProcessedDirs) {
      // check to see if this sub-directory already has a 'pics.htm' file:
      for (int i=0; i<num; i++) {
        if (entries[i].equals("pics.htm")) {
          outputText.append(directory + " already has a 'pics.htm' file" + nl);
          return;
        }
      }
    }

    // Open HTML output file:
    String fname=directory + fs + "pics.htm";
    File out_file = new File(fname);
    try {
      if (out_file.exists())  {
        try { out_file.delete(); }
        catch (SecurityException e) {  }
      }
    } catch (Exception e) { }

    DataOutputStream outs = null;
    try {
      //System.out.println("Opening file " + fname + " for writing.");
      try {
        File d_f = new File(fname);
        outs = new DataOutputStream(new FileOutputStream(d_f));
      } catch (Exception e) {
        outputText.append("Exception writing HTML file: " + e + nl);
        return;
      }

      outs.writeBytes("<h1>Directory " + directory.replace('\\','/') + "</h1>");

      // Make 2 passes at these files:
      //   1. write out HTML aref's to sub-directories
      //   2. write out HTML aref's showing mini-files, linked to large graphics files

      String [] entry_names = new String [500];

      if (num>499) num=499;
      for (int i=0; i<num; i++) {
        entry_names[i] = entries[i].toString();
      }

      for(int i = 0; i <num; i++) {
        try {
          File f = new File(cwd, entry_names[i]);
          if (f.isDirectory()) {
            // This is a directory, so add an entry to current HTML file:
            outs.writeBytes("<a href=\""  + entry_names[i].replace('\\','/') +
                            "/pics.htm" +
                            "\">" + directory.replace('\\','/') + 
                            "/" + entry_names[i].replace('\\','/') +
                            "</a><p>");
            // Recurse down to handle this directory: (depth first is OK here!)
            ProcessGifs(directory +  fs  + entry_names[i]);
          } 
        }
        catch (Exception e) { }
      }
      for(int i = 0; i <num; i++) {
        try {
          File f = new File(cwd, entry_names[i]);  // toString OK here ??
          if (!f.isDirectory() && entry_names[i].equals("pics.htm")==false) {

            String sb = 
              (entry_names[i].substring(0,
                                        entry_names[i].length()-4)) +
              "_s" + ".gif";
                
            // This is a graphics file, so shrink the file:
            if (output_image != null) output_image.flush();
            output_image=newImage(directory + fs + entry_names[i], 1);
            // Add a link:
            if (output_image!=null) {
              repaint();
              if (fileNamesFlag) {
                outs.writeBytes("  " + entry_names[i].replace('\\','/') + " ");
              }
              outs.writeBytes("<a href=\"" + entry_names[i].replace('\\','/') + "\"> <IMG src=\"" +
                              sb.replace('\\', '/') + "\" border=0></a>");

            }
          }
        } 
        catch (Exception e) { 
        }
      }
    }
    finally {
      outs.close();
    }
  }

  final private Image newImage(String file_name, int scale_mode) {

    File test_file = new File(file_name);
    if (test_file.isDirectory()) return null;

    int
      to_width, to_height;
    int
      image_width, image_height,
      image_pixels[];

    // read data 

    URL url=null;

    try { url = new URL("file", "", file_name); }
    catch (MalformedURLException e) { }
    if (url==null) { return null; }

    StringBuffer output_name = 
      new StringBuffer(file_name.substring(0, file_name.length()-4));
    output_name.append("_s.gif");

    Image input_image = java.awt.Toolkit.getDefaultToolkit().getImage(url);

    MediaTracker
      tracker = new MediaTracker(this);
    tracker.addImage(input_image, 0);
    // load it all before continuing
    try { tracker.waitForAll(); } catch (InterruptedException e){};
    tracker = null;

    image_width = input_image.getWidth(this);
    image_height = input_image.getHeight(this);

    image_pixels = new int[image_width * image_height];

    to_width = 60;
    to_height = 60;

    float fact2;
    if (image_width>image_height) {
      fact2=image_width;
    } else {
      fact2=image_height;
    }

    fact2 = 60 / fact2;

    to_width = (int)((float)image_width * fact2);
    to_height= (int)((float)image_height* fact2);

    // Tweak the size of the thumbnails:
    if (imageSizeChoice == 0) {
      to_width = (2 * to_width) / 3;
      to_height= (2 * to_height)/ 3;
    }
    if (imageSizeChoice == 2) {
      to_width = (3 * to_width) / 2;
      to_height= (3 * to_height)/ 2;
    }

    //# Create a PixelGrabber to Get the Pixels of the image and store
    //# them into the image_pixels array
    PixelGrabber pixel_grabber =
      new PixelGrabber(input_image.getSource(), 0, 0,
                       image_width, image_height, image_pixels,
                       0, image_width);

    try {
      pixel_grabber.grabPixels();
    } catch (InterruptedException e) { }
    pixel_grabber = null;

    // resample data
    int 
      to_pixels[] = new int[to_width*to_height];
    double
      xscale, yscale;

    // set scale
    xscale = ((double)image_width) / ((double)to_width);
    yscale = ((double)image_height) / ((double)to_height);

    // check mode
    if ( scale_mode != 0 ) {   // 1 to 1 scaling
      xscale = java.lang.Math.min(xscale,yscale);
      yscale = xscale;
    }

    int
      from_x, from_y,
      to_index, to_x, to_y;

    for(to_index=0,to_y=0;to_y<to_height;to_y++) {
      from_y = (int)(yscale * to_y);
      for(to_x=0;to_x<to_width;to_x++,to_index++) {
        from_x = (int)(xscale * to_x);
        to_pixels[to_index] = image_pixels[ (from_y * image_width) + from_x];
      }// end for x
    }// end for y

    // TRY TO MAP IMAGE COLORS DOWN TO ONLY 256 COLORS:
    int size = to_width * to_height;
    int skip = size / 255;

    int num_l = 64;
    for (int ii=0; ii<num_l; ii++) {
      //s_pixels[ii] = to_pixels[ii*skip];
      float xxx = (float)((size - 1) * Math.random());
      int ind = (int)xxx;
      s_pixels[ii] = to_pixels[ind];
      reds[ii] = (s_pixels[ii] >> 16) & 0xff;
      greens[ii] = (s_pixels[ii] >> 8) & 0xff;
      blues[ii] = s_pixels[ii] & 0xff;
      trans[ii] = (s_pixels[ii] >>>24) < 0x80;
    }
    // Now, for each pixel in to_pixels[], find the
    // closest pixel in s_pixels[], and replace the
    // value in the to_pixels[] array:

    for (int j=0; j<size; j++) {
      // just try to find a close match:
      float best_error = 9999999.9f;
      float error = 0;
      int best_index = 0;
      float x, z;
      boolean trans2;
      for (int i=0; i<num_l; i++) {
        error=0.0f;
        x=(float)((to_pixels[j] >> 16) & 0xff);
        z=reds[i];
        error += Math.sqrt((x - z) * (x - z));
        x=(float)((to_pixels[j] >> 8) & 0xff);
        z=greens[i];
        error += Math.sqrt((x - z) * (x - z));
        x=(float)(to_pixels[j] & 0xff);
        z=blues[i];
        error += Math.sqrt((x - z) * (x - z));
        trans2 = (to_pixels[j] >>> 24) < 0x80;
        if (trans[i] != trans2) error += 99999.0f;
        if (error < best_error) {
          best_error = error;
          best_index = i;
        }
      }

      if (num_l < 255 && best_error > 30.0f) {
        error=0.0f;
        s_pixels[num_l] = to_pixels[j];
        reds[num_l] = ((s_pixels[num_l] >> 16) & 0xff);
        greens[num_l] = ((s_pixels[num_l] >> 8) & 0xff);
        blues[num_l] = (s_pixels[num_l] & 0xff);
        trans[num_l] = (s_pixels[num_l] >>>24) < 0x80;
        num_l++;
        best_index=num_l-1;
        x=(float)(((to_pixels[j] >> 16) & 0xff));
        z=(float)(reds[best_index]);
        error += Math.sqrt((x - z) * (x - z));
        x=(float)(((to_pixels[j] >> 8) & 0xff));
        z=(float)(greens[best_index]);
        error += Math.sqrt((x - z) * (x - z));
        x=(float)((to_pixels[j] & 0xff));
        z=(float)(blues[best_index]);
        error += Math.sqrt((x - z) * (x - z));
        trans2 = (to_pixels[j] >>> 24) < 0x80;
        if (trans[best_index] != trans2) error += 99999.0f;
        best_error=error;
      }
      to_pixels[j]=s_pixels[best_index];
    }

    // new image
    Image temp= createImage(new MemoryImageSource(to_width,to_height,
                                                  to_pixels,0,to_width));

    if (showLogo) {
      logo.setImage(temp, to_width, to_height);
    }
      
    // save the image to a new GIF file:
    FileOutputStream os = null;
    try {
      StringBuffer new_name =
        new StringBuffer((new String(output_name)).substring(0, output_name.length()-4));
      new_name.append(".gif");
      File out_file = new File(new String(new_name));
      os = new FileOutputStream(out_file);
      GifEncoder ge = new GifEncoder(temp, os);
      ge.encode();
      temp = null;

      try { os.close(); } catch (IOException e) { 
      }
    } catch (Exception e) { 
      try { os.close(); } catch (IOException e2) {
      }
    }
    return input_image;  // temp;
  }

  public void paint(Graphics g) {
  }

  private transient TextField root_dir;
  private transient TextArea outputText;
  private boolean process;
  private Thread workThread;

  private class GraphicsFileFilter implements FilenameFilter {
    public GraphicsFileFilter() {
    }
    final public boolean accept(File dir, String name) {
      if (name.endsWith("_s.gif")) return false;
      if (name.endsWith("_S.gif")) return false;
      if (name.endsWith("gif")) return true;
      if (name.endsWith("jpg")) return true;
      if (name.endsWith("bmp")) return true;
      if (name.endsWith("GIF")) return true;
      if (name.endsWith("JPG")) return true;
      if (name.endsWith("BMP")) return true;
      if (name.endsWith("htm")) return true;
      if (name.endsWith("HTM")) return true;
      return (new File(dir, name)).isDirectory();
    }
  }
  String copyright = 
  "PicWeb(tm) Copyright 1997 by Mark Watson." + nl +
  "The GifEncoder class is copyright 1996 by" + nl +
  "Jef Poskanzer <jef@acme.com>." + nl + nl +
  "READ FULL COPYRIGHT NOTICES ON OPTIONS SCREEN" + nl;

  /** class Options
  */
  
  private class Options extends Dialog implements Serializable {
    private Button close;
    private TextArea text;
    private DialogFrame my_frame;

    private Checkbox cbShowLogo;
    private Checkbox cbFileNames;
    private Checkbox cbSkipDirs;
    private Choice choice;

    Options(DialogFrame parent) {
      super(parent, "Options dialog", true);
      my_frame = parent;
      my_frame.setLayout(null);
      my_frame.setOptions(this);
      
      cbShowLogo = new Checkbox("Show images while calculated", true);
      cbFileNames= new Checkbox("Include image file names in HTML documents", false);
      cbSkipDirs = new Checkbox("Skip already processed sub-directories", false);

      Label l1 = new Label("Set options");
      l1.setFont(new Font("Dialog", Font.BOLD, 20));
      my_frame.add(l1);
      l1.setBounds(5, 180, 480, 28);

      my_frame.add(cbShowLogo);
      cbShowLogo.setBounds(5, 210, 480, 28);
      my_frame.add(cbFileNames);
      cbFileNames.setBounds(5, 240, 480, 28);
      my_frame.add(cbSkipDirs);
      cbSkipDirs.setBounds(5, 270, 480, 28);

      choice = new Choice();
      choice.add("generate small size thumbnail images");
      choice.add("generate medium size thumbnail images");
      choice.add("generate large size thumbnail images");
      choice.select(2);
      my_frame.add(choice);
      choice.setBounds(5, 310, 480, 28);

      String str =
        "PicWeb(tm) Copyright 1997 by Mark Watson." + nl + nl + "ALL RIGHTS RESERVED" + nl + "Single User License." + nl +
        "The GifEncoder class is copyright:" + nl +
        "Copyright (C) 1996 by Jef Poskanzer <jef@acme.com>.  All rights reserved." + nl + nl +
        "THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND" + nl +
        "ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE" + nl +
        "IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE" + nl +
        "ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE" + nl +
        "FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL" + nl +
        "DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS" + nl +
        "OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)" + nl +
        "HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT" + nl +
        "LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY" + nl +
        "OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF" + nl +
        "SUCH DAMAGE." + nl + nl +
        "Visit my We site for update information: www.markwatson.com" + nl +
        "Thanks to Jef Poskanzer for the GifEncoder class.  Jef's web site" + nl +
        "can be found at: www.acme.com" + fs + "java" + nl + nl +
        "NOTICE: THIS SOFTWARE CAN NOT BE USED FOR ANY ILLEGAL OR NON-PEACEFUL PURPOSE" + nl + nl +
        "-----------------------------------------------------------------------" + nl + nl +
        "HELP" + nl + nl +
        "Use the forward slash: / as a file separator.  If possible, use a Java" + nl +
        "runtime system with a just in time compiler to decrease the runtime for" + nl +
        "generating many thumbnail image files." + nl;

      Label l = new Label("Copyright information and help");
      l.setFont(new Font("Dialog", Font.BOLD, 24));
      my_frame.add(l);
      l.setBounds(5, 25, 480, 30);

      text = new TextArea(str, 7, 72);
      text.setEditable(false);
      text.setFont(new Font("Dialog", Font.PLAIN, 10));
      text.setEditable(false);
      my_frame.add(text);
      text.setBounds(5, 60, 480, 110);
      
      Label l9 = new Label("Close this window to save options!");
      l9.setFont(new Font("Dialog", Font.BOLD, 20));
      my_frame.add(l9);
      l9.setBounds(5, 350, 480, 25);
      
      my_frame.setBounds(100, 100, 490, 390);
      my_frame.pack();
      myself=this;
    }

    public void setVisible(boolean b) {
      my_frame.setVisible(b);
    }

    private Options myself;

    protected void saveOptions() {
      showLogo = cbShowLogo.getState();
      fileNamesFlag = cbFileNames.getState();
      skipProcessedDirs = cbSkipDirs.getState();
      imageSizeChoice = choice.getSelectedIndex();
      myself.setVisible(false);
    }
  }

  class DialogFrame extends Frame {
    public DialogFrame() {
      super(); 
      enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    }
    public void setOptions(Options o) {
      options = o;
    }
    private Options options;
    public void processEvent(AWTEvent evt) { // JDK 1.1
      switch (evt.getID())  {
      case Event.WINDOW_DESTROY:
        options.saveOptions();
      default:
        super.processEvent(evt);
      }                  
    }  
  }
}
